﻿using Firebase;
using Firebase.Auth;
using Firebase.Database;
using System;
using System.Collections;
using System.Linq;
using TMPro;
using UnityEngine;


//the script for authentication and updating results of users
public class FirebaseManager : MonoBehaviour
{
    //Variables of Firebase
    public DependencyStatus dependencyStatus;
    public DatabaseReference DatabaseReference;
    public FirebaseUser FirebaseUser;
    public FirebaseAuth FirebaseAuth;


    //Variables used in "Register" window
    public TMP_InputField RegisterInputUsernameField;
    public TMP_InputField RegisterInputEmailField;
    public TMP_InputField RegisterInputPasswordField;
    public TMP_InputField RegisterInputPasswordVerifyField;
    public TMP_Text RegisterWarningText;


    //Variables used in "Login" window
    public TMP_InputField LoginInputEmailField;
    public TMP_InputField LoginInputPasswordField;
    public TMP_Text LoginConfirmedText;
    public TMP_Text LoginWarningText;


    //Variables used in "User Data" window
    public TMP_InputField UserDataUsernameField;
    public TMP_InputField UserDataBestSolveField;
    public TMP_InputField UserDataAvg5Field;
    public TMP_InputField UserDataAvg12Field;
    public GameObject SolvingElement;
    public Transform TableOfUsersContent;


    //Function to check if all the necessary dependencies for Firebase Database are present and
    //if they are, then Firebase Database will be set
    void Awake()
    {
        FirebaseApp.CheckAndFixDependenciesAsync().ContinueWith(task =>
        {
            dependencyStatus = task.Result;
            if (dependencyStatus == DependencyStatus.Available)
                SettingFirebaseDatabase();
        });
    }

    private void SettingFirebaseDatabase()
    {
        FirebaseAuth = FirebaseAuth.DefaultInstance;
        DatabaseReference = FirebaseDatabase.DefaultInstance.RootReference;
    }

    //Function to clear Register Fields
    public void RegisterFieldsClean()
    {
        RegisterInputUsernameField.text = "";
        RegisterInputEmailField.text = "";
        RegisterInputPasswordField.text = "";
        RegisterInputPasswordVerifyField.text = "";
    }

    //Function to clear Login Fields
    public void LoginFieldsClean()
    {
        LoginInputEmailField.text = "";
        LoginInputPasswordField.text = "";
    }


    //Function when "Login" button was clicked
    public void ClickOnLoginButton()
    {
        StartCoroutine(LoginProcess(LoginInputEmailField.text, LoginInputPasswordField.text));
    }
    //Function when "Register" button was clicked
    public void ClickOnRegisterButton()
    {
        StartCoroutine(RegisterProcess(RegisterInputEmailField.text, RegisterInputPasswordField.text, RegisterInputUsernameField.text));
    }

    //Function when "Save" button was clicked
    public void ClickOnSaveButton()
    {
        StartCoroutine(UpdateUsername(UserDataUsernameField.text));
        StartCoroutine(UpdateUsernameDatabase(UserDataUsernameField.text));
        StartCoroutine(Updatebestsolve(float.Parse(UserDataBestSolveField.text)));
        StartCoroutine(Updateavg5(float.Parse(UserDataAvg5Field.text)));
        StartCoroutine(Updateavg12(float.Parse(UserDataAvg12Field.text)));
    }
    //Function when "Sign Out" button was clicked
    public void ClickOnSignOutButton()
    {
        FirebaseAuth.SignOut();
        UIManager.window.LoginWindowOnClick();
        RegisterFieldsClean();
        LoginFieldsClean();
    }

    //Function for the scoreboard button
    public void ClickOnTableOfUsersButton()
    {
        StartCoroutine(LoadTableOfUsers());
    }

    //Function of Login Process that verifies if the user exists in the system
    private IEnumerator LoginProcess(string _emailValue, string _passwordValue)
    {
        var LoginTask = FirebaseAuth.SignInWithEmailAndPasswordAsync(_emailValue, _passwordValue);
        yield return new WaitUntil(predicate: () => LoginTask.IsCompleted);

        if (LoginTask.Exception != null)
        {
            //If some error occurs, the system will display the message with corresponding text of the error
            FirebaseException FirebaseException = LoginTask.Exception.GetBaseException() as FirebaseException;
            AuthError ErrorCode = (AuthError)FirebaseException.ErrorCode;

            string ErrorMessage = "Login Failed!";
            switch (ErrorCode)
            {
                case AuthError.UserNotFound:
                    ErrorMessage = "Account does not exist";
                    break;
                case AuthError.MissingEmail:
                    ErrorMessage = "Missing Email";
                    break;
                case AuthError.InvalidEmail:
                    ErrorMessage = "Invalid Email";
                    break;
                case AuthError.WrongPassword:
                    ErrorMessage = "Wrong Password";
                    break;
                case AuthError.MissingPassword:
                    ErrorMessage = "Missing Password";
                    break;
            }
            LoginWarningText.text = ErrorMessage;
        }
        else
        {
            //Otherwise, a user will be logged in successfully with a green-coloured message "Logged In"
            FirebaseUser = LoginTask.Result;
            LoginConfirmedText.text = "Logged In";
            LoginWarningText.text = "";
            StartCoroutine(LoadUserData());

            yield return new WaitForSeconds(1);

            UserDataUsernameField.text = FirebaseUser.DisplayName;
            UIManager.window.UserDataWindowOnClick();
            LoginConfirmedText.text = "";
            RegisterFieldsClean();
            LoginFieldsClean();
        }
    }

    private IEnumerator RegisterProcess(string _emailValue, string _passwordValue, string _usernameValue)
    {
        if (_usernameValue == "")
        {
            //Warning message will be shown if the user did not input their username
            RegisterWarningText.text = "Missing Username";
        }
        else if (RegisterInputPasswordField.text != RegisterInputPasswordVerifyField.text)
        {
            //Warning message will be shown if passwords do not match
            RegisterWarningText.text = "Password Does Not Match!";
        }
        else
        {
            var RegisterTask = FirebaseAuth.CreateUserWithEmailAndPasswordAsync(_emailValue, _passwordValue);
            yield return new WaitUntil(predicate: () => RegisterTask.IsCompleted);

            if (RegisterTask.Exception != null)
            {
                //If some error occurs, the system will display the message with corresponding text of the error
                FirebaseException FirebaseException = RegisterTask.Exception.GetBaseException() as FirebaseException;
                AuthError ErrorCode = (AuthError)FirebaseException.ErrorCode;

                string ErrorMessage = "Register Failed!";
                switch (ErrorCode)
                {
                    case AuthError.EmailAlreadyInUse:
                        ErrorMessage = "Email Already In Use";
                        break;
                    case AuthError.MissingEmail:
                        ErrorMessage = "Missing Email";
                        break;
                    case AuthError.MissingPassword:
                        ErrorMessage = "Missing Password";
                        break;
                    case AuthError.WeakPassword:
                        ErrorMessage = "Weak Password";
                        break;
                }
                RegisterWarningText.text = ErrorMessage;
            }
            else
            {
                FirebaseUser = RegisterTask.Result;

                if (FirebaseUser != null)
                {
                    //New user is created with a specific username
                    UserProfile userProfile = new UserProfile { DisplayName = _usernameValue };
                    var ProfileTask = FirebaseUser.UpdateUserProfileAsync(userProfile);
                    yield return new WaitUntil(predicate: () => ProfileTask.IsCompleted);

                    if (ProfileTask.Exception != null)
                    {
                        RegisterWarningText.text = "Username Set Failed!";
                    }
                    else
                    {
                        //After successful registration, "Login" page is opened
                        UIManager.window.LoginWindowOnClick();
                        RegisterWarningText.text = "";
                        LoginFieldsClean();
                        RegisterFieldsClean();

                    }
                }
            }
        }
    }
    private IEnumerator UpdateUsername(string _username)
    {
        UserProfile profile = new UserProfile { DisplayName = _username };
        var ProfileTask = FirebaseUser.UpdateUserProfileAsync(profile);
        yield return new WaitUntil(predicate: () => ProfileTask.IsCompleted);
    }

    //Updates the value of "Username" in Firebase Database
    private IEnumerator UpdateUsernameDatabase(string _usernameValue)
    {
        //Set the currently logged in user username in the database
        var DatabaseTask = DatabaseReference.Child("users").Child(FirebaseUser.UserId).Child("username").SetValueAsync(_usernameValue);
        yield return new WaitUntil(predicate: () => DatabaseTask.IsCompleted);
    }

    //Updates the value of "Best Solve" in Firebase Database
    private IEnumerator Updatebestsolve(float _bestSolveValue)
    {
        //Set the currently logged in user bestsolve
        var DatabaseTask = DatabaseReference.Child("users").Child(FirebaseUser.UserId).Child("bestsolve").SetValueAsync(_bestSolveValue);
        yield return new WaitUntil(predicate: () => DatabaseTask.IsCompleted);
    }

    //Updates the value of "Average of 5" in Firebase Database
    private IEnumerator Updateavg5(float _avg5Value)
    {
        //Set the currently logged in user avg5
        var DatabaseTask = DatabaseReference.Child("users").Child(FirebaseUser.UserId).Child("avg5").SetValueAsync(_avg5Value);
        yield return new WaitUntil(predicate: () => DatabaseTask.IsCompleted);
    }
    //Updates the value of "Average of 12" in Firebase Database
    private IEnumerator Updateavg12(float _avg12Value)
    {
        //Set the currently logged in user avg12
        var DatabaseTask = DatabaseReference.Child("users").Child(FirebaseUser.UserId).Child("avg12").SetValueAsync(_avg12Value);
        yield return new WaitUntil(predicate: () => DatabaseTask.IsCompleted);
    }

    //Loads User Data
    private IEnumerator LoadUserData()
    {
        //Get the currently logged in user data
        var DatabaseTask = DatabaseReference.Child("users").Child(FirebaseUser.UserId).GetValueAsync();
        yield return new WaitUntil(predicate: () => DatabaseTask.IsCompleted);

        if (DatabaseTask.Result.Value == null)
        {
            //No data yet
            UserDataBestSolveField.text = "0";
            UserDataAvg5Field.text = "0";
            UserDataAvg12Field.text = "0";
        }
        else
        {
            //Retrives the data
            DataSnapshot snapshot = DatabaseTask.Result;
            UserDataBestSolveField.text = snapshot.Child("bestsolve").Value.ToString();
            UserDataAvg5Field.text = snapshot.Child("avg5").Value.ToString();
            UserDataAvg12Field.text = snapshot.Child("avg12").Value.ToString();
        }
    }

    //Loads Table of Users
    private IEnumerator LoadTableOfUsers()
    {
        //Sorts user data by the value of "Best Solve"
        var DatabaseTask = DatabaseReference.Child("users").OrderByChild("bestsolve").GetValueAsync();
        yield return new WaitUntil(predicate: () => DatabaseTask.IsCompleted);



        DataSnapshot dataSnapshot = DatabaseTask.Result;

        //Disables all current elements of Table of Users
        foreach (Transform child in TableOfUsersContent.transform)
        {
            Destroy(child.gameObject);
        }

        foreach (DataSnapshot databasevalue in dataSnapshot.Children.Reverse<DataSnapshot>())
        {
            string usernameValue = databasevalue.Child("username").Value.ToString();
            float bestSolveValue = float.Parse(databasevalue.Child("bestsolve").Value.ToString());
            float avg5Value = float.Parse(databasevalue.Child("avg5").Value.ToString());
            float avg12Value = float.Parse(databasevalue.Child("avg12").Value.ToString());

            GameObject tableElement = Instantiate(SolvingElement, TableOfUsersContent);
            tableElement.GetComponent<TableOfUsers>().NewScoreElement(usernameValue, bestSolveValue, avg5Value, avg12Value);
        }
        //Loads "Table of Users" window
        UIManager.window.TableOfUsersWindowOnClick();

    }
}